package umlproblemanalysis.popup.actions;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.common.util.EList;
import org.eclipse.uml2.uml.Activity;
import org.eclipse.uml2.uml.ActivityEdge;
import org.eclipse.uml2.uml.ActivityNode;
import org.eclipse.uml2.uml.DecisionNode;
import org.eclipse.uml2.uml.MergeNode;
import org.eclipse.uml2.uml.Stereotype;
import org.eclipse.uml2.uml.StructuredActivityNode;

public class ProblemAnalysis {

	public static ProblemReport perform(List<Activity> activities) {

		List<Problem> problems= new ArrayList<Problem>();

		for (Activity activity : activities) {
			for (ActivityNode activityNode : getInitialNodes(activity.getNodes())) {
				visit(activity, activityNode, new ArrayList<ActivityNode>(), problems);
			}
		}

		return new ProblemReport(problems);
	}

	private static void visit(Activity activity, ActivityNode activityNode, List<ActivityNode> visitedNodes, List<Problem> problems) {

		if (visitedNodes.contains(activityNode)) {
			// cycle.
			if (! (activityNode instanceof DecisionNode)) {
				// Log a problem.
				problems.add(new Problem(activity, activityNode, visitedNodes, Problem.ProblemCode.RETURNNOTTODECISION));
			}
			// Do not continue
			return;
		}

		// Structured activity node?
		if (activityNode instanceof StructuredActivityNode) {
			StructuredActivityNode san= ((StructuredActivityNode) activityNode);
			EList<ActivityNode> containedNodes= san.getContainedNodes();

			for (ActivityNode containedNode : getInitialNodes(containedNodes)) {
				List<ActivityNode> current= new ArrayList<ActivityNode>();
				current.add(san);
				visit(activity, containedNode, current, problems);
			}

		}


		List<ActivityEdge> outgoings= getFilteredEdges(activityNode.getOutgoings());
		if (outgoings.size() > 1 && ! (activityNode instanceof DecisionNode) && ! (activityNode instanceof MergeNode))
			problems.add(new Problem(activity, activityNode, visitedNodes, Problem.ProblemCode.WRONGCONTROLFLOWSPLIT));

		for (ActivityEdge activityEdge : outgoings) {
			ActivityNode target= activityEdge.getTarget();

			List<ActivityEdge> incomings= getFilteredEdges(target.getIncomings());
			if (incomings.size() > 1 && ! (target instanceof DecisionNode) && ! (target instanceof MergeNode))
				problems.add(new Problem(activity, target, visitedNodes, Problem.ProblemCode.WRONGCONTROLFLOWMERGE));

			visitedNodes.add(activityNode);
			visit(activity, target, visitedNodes, problems);
			visitedNodes.remove(activityNode);
		}
	}

	private static List<ActivityEdge> getFilteredEdges(EList<ActivityEdge> outgoings) {
		List<ActivityEdge> edges= new ArrayList<ActivityEdge>();
		for (ActivityEdge activityEdge : outgoings) {
			EList<Stereotype> appliedStereotypes= activityEdge.getAppliedStereotypes();
			boolean add= true;
			for (Stereotype stereotype : appliedStereotypes) {
				if (stereotype.getName().equals("exception") || stereotype.getName().equals("compensation") | stereotype.getName().equals("event")) {
					add= false;
					break;
				}
			}
			if (add)
				edges.add(activityEdge);
		}
		return edges;
	}

	private static List<ActivityNode> getInitialNodes(EList<ActivityNode> nodes) {
		List<ActivityNode> result= new ArrayList<ActivityNode>();
		for (ActivityNode activityNode : nodes) {
			if (activityNode.getIncomings().isEmpty())
				result.add(activityNode);
		}
		return result;
	}

}
